/*
 * sch_sfb.c    Stochastic Fair Blue
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Juliusz Chroboczek <jch@pps.jussieu.fr>
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/skbuff.h>
#include <linux/random.h>
#include <linux/jhash.h>
#include <net/ip.h>
#include <net/pkt_sched.h>
#include <net/inet_ecn.h>

#include "sfb.h"
#include "sfb_compat.h"

#define NUMHASHES 10
#define NUMBUCKETS 32

struct bucket {
        u32 qlen;
        u16 pm;
};

struct sfb_sched_data
{
        u32 limit, max, target;
        u16 increment, decrement;
        u32 penalty_rate, penalty_burst;
        u32 bytes_avail;
        psched_time_t token_time;

        u32 perturbation[NUMHASHES];
        struct bucket buckets[NUMHASHES][NUMBUCKETS];
	struct Qdisc *qdisc;

        __u32 earlydrop, penaltydrop, bucketdrop, queuedrop, marked;
};

static unsigned sfb_hash(struct sk_buff *skb, int perturbation)
{
	u32 h, h2;

	switch (skb->protocol) {
	case __constant_htons(ETH_P_IP):
	{
		const struct iphdr *iph = ip_hdr(skb);
		h = iph->daddr;
		h2 = iph->saddr^iph->protocol;
		if (!(iph->frag_off&htons(IP_MF|IP_OFFSET)) &&
		    (iph->protocol == IPPROTO_TCP ||
		     iph->protocol == IPPROTO_UDP ||
		     iph->protocol == IPPROTO_UDPLITE ||
		     iph->protocol == IPPROTO_SCTP ||
		     iph->protocol == IPPROTO_DCCP ||
		     iph->protocol == IPPROTO_ESP))
			h2 ^= *(((u32*)iph) + iph->ihl);
		break;
	}
	case __constant_htons(ETH_P_IPV6):
	{
		struct ipv6hdr *iph = ipv6_hdr(skb);
		h = iph->daddr.s6_addr32[3];
		h2 = iph->saddr.s6_addr32[3]^iph->nexthdr;
		if (iph->nexthdr == IPPROTO_TCP ||
		    iph->nexthdr == IPPROTO_UDP ||
		    iph->nexthdr == IPPROTO_UDPLITE ||
		    iph->nexthdr == IPPROTO_SCTP ||
		    iph->nexthdr == IPPROTO_DCCP ||
		    iph->nexthdr == IPPROTO_ESP)
			h2 ^= *(u32*)&iph[1];
		break;
	}
	default:
		h = (u32)(unsigned long)skb->dst^skb->protocol;
		h2 = (u32)(unsigned long)skb->sk;
	}

        return jhash_2words(h, h2, perturbation) % NUMBUCKETS;

}

static inline u16 prob_plus(u16 p1, u16 p2)
{
        return p1 < SFB_MAX_PROB - p2 ? p1 + p2 : SFB_MAX_PROB;
}

static inline u16 prob_minus(u16 p1, u16 p2)
{
        return p1 > p2 ? p1 - p2 : 0;
}

static void
increment_qlen(struct sk_buff *skb, struct sfb_sched_data *q)
{
        int i;
        for(i = 0; i < NUMHASHES; i++) {
                int hash = sfb_hash(skb, q->perturbation[i]);
                q->buckets[i][hash].qlen++;
        }
}

static void
decrement_qlen(struct sk_buff *skb, struct sfb_sched_data *q)
{
        int i;
        for(i = 0; i < NUMHASHES; i++) {
                int hash = sfb_hash(skb, q->perturbation[i]);
                if(q->buckets[i][hash].qlen > 0)
                    q->buckets[i][hash].qlen--;
        }
}

static void
zero_all_buckets(struct sfb_sched_data *q)
{
        int i, j;
        for(i = 0; i < NUMHASHES; i++) {
                for(j = 0; j < NUMBUCKETS; j++) {
                        q->buckets[i][j].pm = 0;
                        q->buckets[i][j].qlen = 0;
                }
        }
}

static void
compute_qlen(u32 *qlen_r, u16 *prob_r, struct sfb_sched_data *q)
{
        int i, j;
        u32 qlen = 0;
        u16 prob = 0;
        for(i = 0; i < NUMHASHES; i++) {
                for(j = 0; j < NUMBUCKETS; j++) {
                        if(qlen < q->buckets[i][j].qlen)
                                qlen = q->buckets[i][j].qlen;
                        if(qlen < q->buckets[i][j].pm)
                                prob = q->buckets[i][j].pm;
                }
        }
        *qlen_r = qlen;
        *prob_r = prob;
}


static void
init_perturbation(struct sfb_sched_data *q)
{
        get_random_bytes(q->perturbation, sizeof(q->perturbation));
}

static int rate_limit(struct sk_buff *skb, struct sfb_sched_data* q)
{
        if(q->penalty_rate == 0 || q->penalty_burst == 0)
                return 1;

        if(q->bytes_avail < skb->len) {
                psched_time_t now;
                psched_tdiff_t age;

                now = psched_get_time();
                age = psched_tdiff_bounded(now, q->token_time, 0x80000000);
                q->bytes_avail =
                        (age * q->penalty_rate / PSCHED_TICKS_PER_SEC);
                if(q->bytes_avail > q->penalty_burst)
                        q->bytes_avail = q->penalty_burst;
                q->token_time = now;
                if(q->bytes_avail < skb->len)
                        return 1;
        }

        q->bytes_avail -= skb->len;
        return 0;
}

static int sfb_enqueue(struct sk_buff *skb, struct Qdisc* sch)
{

	struct sfb_sched_data *q = qdisc_priv(sch);
        struct Qdisc *child = q->qdisc;
        u16 minprob = SFB_MAX_PROB;
        u32 minqlen = (u32)(~0);
        int ret, i;

        for(i = 0; i < NUMHASHES; i++) {
                int hash = sfb_hash(skb, q->perturbation[i]);
                if(unlikely(q->buckets[i][hash].qlen >= q->target)) {
                    q->buckets[i][hash].pm =
                        prob_plus(q->buckets[i][hash].pm, q->increment);
                } else if(q->buckets[i][hash].qlen == 0) {
                    q->buckets[i][hash].pm =
                        prob_minus(q->buckets[i][hash].pm, q->decrement);
                }
                if(minqlen > q->buckets[i][hash].qlen)
                        minqlen = q->buckets[i][hash].qlen;
                if(minprob > q->buckets[i][hash].pm)
                        minprob = q->buckets[i][hash].pm;
        }
        
        if(unlikely(minqlen >= q->max || sch->q.qlen >= q->limit)) {
                sch->qstats.overlimits++;
                if(sch->q.qlen >= q->limit)
                        q->queuedrop++;
                else
                        q->bucketdrop++;
                goto drop;
        } else if(minprob >= SFB_MAX_PROB) {
                if(rate_limit(skb, q)) {
                        sch->qstats.overlimits++;
                        q->penaltydrop++;
                        goto drop;
                }
        } else if((net_random() & SFB_MAX_PROB) < minprob) {
                if(!INET_ECN_set_ce(skb)) {
                        q->earlydrop++;
                        goto drop;
                } else {
                        q->marked++;
                }
        }

        ret = child->enqueue(skb, child);
        if(likely(ret == NET_XMIT_SUCCESS)) {
                sch->q.qlen++;
                increment_qlen(skb, q);
		sch->bstats.packets++;
		sch->bstats.bytes += skb->len;
                sch->qstats.backlog += skb->len;
        } else {
                q->queuedrop++;
        }
        return ret;

 drop:
        qdisc_drop(skb, sch);
        return NET_XMIT_CN;
}

static int sfb_requeue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
        struct Qdisc *child = q->qdisc;
        int ret;

        ret = child->ops->requeue(skb, child);
        if(unlikely(ret != NET_XMIT_SUCCESS)) {
                sch->qstats.drops++;
                return ret;
        }

        sch->q.qlen++;
        increment_qlen(skb, q);
        sch->qstats.backlog += skb->len;
        sch->qstats.requeues++;
        return ret;
}

static struct sk_buff *sfb_dequeue(struct Qdisc* sch)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
        struct Qdisc *child = q->qdisc;
	struct sk_buff *skb;

	skb = child->dequeue(q->qdisc);

        if(skb) {
                sch->q.qlen--;
		sch->qstats.backlog -= skb->len;
                decrement_qlen(skb, q);
        }

        return skb;
}

static unsigned int sfb_drop(struct Qdisc* sch)
{
        /* Since child->drop doesn't return the dropped skb, we have no
           way to find out the buckets to update. */
        return 0;
}

static void sfb_reset(struct Qdisc* sch)
{
	struct sfb_sched_data *q = qdisc_priv(sch);

	qdisc_reset(q->qdisc);
        sch->q.qlen = 0;
        sch->qstats.backlog = 0;
        q->earlydrop = 0;
        q->penaltydrop = 0;
        q->bucketdrop = 0;
        q->queuedrop = 0;
        q->marked = 0;
        zero_all_buckets(q);
        init_perturbation(q);
}

static void sfb_destroy(struct Qdisc *sch)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
	qdisc_destroy(q->qdisc);
        q->qdisc = NULL;
}

static struct Qdisc *sfb_create_dflt(struct Qdisc *sch, u32 limit)
{
	struct Qdisc *q;
	struct rtattr *rta;
	int ret;

	q = qdisc_create_dflt(sch->dev, &pfifo_qdisc_ops,
			      TC_H_MAKE(sch->handle, 1));
	if (q) {
		rta = kmalloc(RTA_LENGTH(sizeof(struct tc_fifo_qopt)),
			      GFP_KERNEL);
		if (rta) {
			rta->rta_type = RTM_NEWQDISC;
			rta->rta_len = RTA_LENGTH(sizeof(struct tc_fifo_qopt));
			((struct tc_fifo_qopt *)RTA_DATA(rta))->limit = limit;

			ret = q->ops->change(q, rta);
			kfree(rta);

			if (ret == 0)
				return q;
		}
		qdisc_destroy(q);
	}
	return NULL;
}

static int sfb_change(struct Qdisc *sch, struct rtattr *opt)
{
        struct sfb_sched_data *q = qdisc_priv(sch);
	struct Qdisc *child = NULL;
	struct rtattr *tb[TCA_SFB_MAX];
	struct tc_sfb_qopt *ctl;
        u32 limit, max, target;
        u16 increment, decrement;
        u32 penalty_rate, penalty_burst;

	if (opt == NULL) {
                limit = 0;
                max = 30;
                target = 20;
                increment = (SFB_MAX_PROB + 1000) / 2000;
                decrement = (SFB_MAX_PROB + 10000) / 20000;
                penalty_rate = 6000;
                penalty_burst = 12000;
        } else {
                if(rtattr_parse_nested(tb, TCA_SFB_MAX, opt))
                        return -EINVAL;

                if (tb[TCA_SFB_PARMS-1] == NULL ||
                    RTA_PAYLOAD(tb[TCA_SFB_PARMS-1]) < sizeof(*ctl))
                        return -EINVAL;
                
                ctl = RTA_DATA(tb[TCA_SFB_PARMS-1]);
                limit = ctl->limit;
                max = ctl->max;
                target = ctl->target;
                increment = ctl->increment;
                decrement = ctl->decrement;
                penalty_rate = ctl->penalty_rate;
                penalty_burst = ctl->penalty_burst;
        }

        if(limit == 0)
                limit = sch->dev->tx_queue_len ? sch->dev->tx_queue_len : 1;

        child = sfb_create_dflt(sch, limit);
        if(child == NULL)
                return -ENOMEM;
                
        sch_tree_lock(sch);
	if(child) {
		qdisc_tree_decrease_qlen(q->qdisc, q->qdisc->q.qlen);
		qdisc_destroy(xchg(&q->qdisc, child));
	}

        q->limit = limit;
        q->increment = increment;
        q->decrement = decrement;
        q->max = max;
        q->target = target;
        q->penalty_rate = penalty_rate;
        q->penalty_burst = penalty_burst;

        init_perturbation(q);

	sch_tree_unlock(sch);

	return 0;
}

static int sfb_init(struct Qdisc *sch, struct rtattr *opt)
{
      	struct sfb_sched_data *q = qdisc_priv(sch);

        q->qdisc = &noop_qdisc;
        return sfb_change(sch, opt);
}

static int sfb_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
	struct rtattr *opts = NULL;
	struct tc_sfb_qopt opt = { .limit = q->limit,
                                   .max = q->max,
                                   .target = q->target,
                                   .increment = q->increment,
                                   .decrement = q->decrement,
                                   .penalty_rate = q->penalty_rate,
                                   .penalty_burst = q->penalty_burst,
        };

	opts = RTA_NEST(skb, TCA_OPTIONS);
	RTA_PUT(skb, TCA_SFB_PARMS, sizeof(opt), &opt);
	return RTA_NEST_END(skb, opts);

rtattr_failure:
	return -1;
}

static int sfb_dump_stats(struct Qdisc *sch, struct gnet_dump *d)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
	struct tc_sfb_xstats st = { .earlydrop = q->earlydrop,
                                    .penaltydrop = q->penaltydrop,
                                    .bucketdrop = q->bucketdrop,
                                    .marked = q->marked};

        compute_qlen(&st.maxqlen, &st.maxprob, q);

	return gnet_stats_copy_app(d, &st, sizeof(st));
}
        
static int sfb_dump_class(struct Qdisc *sch, unsigned long cl,
			  struct sk_buff *skb, struct tcmsg *tcm)
{
        return -ENOSYS;
}

static int sfb_graft(struct Qdisc *sch, unsigned long arg, struct Qdisc *new,
		     struct Qdisc **old)
{
	struct sfb_sched_data *q = qdisc_priv(sch);

	if (new == NULL)
		new = &noop_qdisc;

	sch_tree_lock(sch);
	*old = xchg(&q->qdisc, new);
	qdisc_tree_decrease_qlen(*old, (*old)->q.qlen);
	qdisc_reset(*old);
	sch_tree_unlock(sch);
	return 0;
}

static struct Qdisc *sfb_leaf(struct Qdisc *sch, unsigned long arg)
{
	struct sfb_sched_data *q = qdisc_priv(sch);
	return q->qdisc;
}

static unsigned long sfb_get(struct Qdisc *sch, u32 classid)
{
	return 1;
}

static void sfb_put(struct Qdisc *sch, unsigned long arg)
{
	return;
}

static int sfb_change_class(struct Qdisc *sch, u32 classid, u32 parentid,
			    struct rtattr **tca, unsigned long *arg)
{
	return -ENOSYS;
}

static int sfb_delete(struct Qdisc *sch, unsigned long cl)
{
	return -ENOSYS;
}

static void sfb_walk(struct Qdisc *sch, struct qdisc_walker *walker)
{
	if (!walker->stop) {
		if (walker->count >= walker->skip)
			if (walker->fn(sch, 1, walker) < 0) {
				walker->stop = 1;
				return;
			}
		walker->count++;
	}
}

static struct tcf_proto **sfb_find_tcf(struct Qdisc *sch, unsigned long cl)
{
	return NULL;
}

static struct Qdisc_class_ops sfb_class_ops =
{
	.graft		=	sfb_graft,
	.leaf		=	sfb_leaf,
	.get		=	sfb_get,
	.put		=	sfb_put,
	.change		=	sfb_change_class,
	.delete		=	sfb_delete,
	.walk		=	sfb_walk,
	.tcf_chain	=	sfb_find_tcf,
	.dump		=	sfb_dump_class,
};

struct Qdisc_ops sfb_qdisc_ops = {
	.id		=	"sfb",
	.priv_size	=	sizeof(struct sfb_sched_data),
	.cl_ops		=	&sfb_class_ops,
	.enqueue	=	sfb_enqueue,
	.dequeue	=	sfb_dequeue,
	.requeue	=	sfb_requeue,
	.drop		=	sfb_drop,
	.init		=	sfb_init,
	.reset		=	sfb_reset,
	.destroy	=	sfb_destroy,
	.change		=	sfb_change,
	.dump		=	sfb_dump,
	.dump_stats	=	sfb_dump_stats,
	.owner		=	THIS_MODULE,
};

static int __init sfb_module_init(void)
{
	return register_qdisc(&sfb_qdisc_ops);
}

static void __exit sfb_module_exit(void)
{
	unregister_qdisc(&sfb_qdisc_ops);
}

module_init(sfb_module_init)
module_exit(sfb_module_exit)

MODULE_DESCRIPTION("Stochastic Fair Blue queue discipline");
MODULE_AUTHOR("Juliusz Chroboczek");
MODULE_LICENSE("GPL");
